// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// This  skeleton file was automatically created by 
//            pkgmaker 0.1
//            Thursday, December 11 1997
// 
// Module: ArrayPixelIterator
// Author: Hassan Siddiqui
//
// Description: 
//
// Class representation of an iterator that handles DAL-based detectormaps.
//
// Modifications: See ChangeLog
// 
// $Id: ArrayPixelIterator.cc,v 1.1 2012/01/25 15:46:31 rsaxton Exp $

// General classes
#include <iostream>
#include<cassert>

// SAS-specific classes
#include "Dal.h"

// General utility functions
#include "errstr.h"
#include "Tracer.h"

// Local classes
#include "ArrayPixelIterator.h"

/* Constructors */
ArrayPixelIterator::ArrayPixelIterator(const string & arrayname) :
  _firstcalled(false),
  _set(0),
  _array(0),
  _arraydata(0),
  // _arrayvalue(0),
  _wcsconv(0)
{

  _set = dataSetServer->
    open( arrayname, 
	  DataSet::Read );

  _array = _set->array(arrayname); 

  _size = _array->elements();
  
  /* Can only handle 2-d arrays */
  if (_array->size().size() !=2)
    {
      errstr << "ArrayPixelIterator : cannot handle an array of size "
	     << _array->size().size()
	     << Sas::error ("invalidArraySize");

      ALTRETURN;
    }

  /* [HS] fixed bug in computation of array dimensions */
  _xbins=(_array->size())[1];

  errstr 
    << "xbins=" << _xbins
    << Sas::message(Msg::App, Msg::Noisy);


  _arraydata = _array->data();

  errstr 
    << "Array datatype " 
    << (long int ) _arraydata->dataType()
    << Sas::message(Msg::App, Msg::Noisy);


  _arrayvalue.data(_arraydata);
  
  /* Create WcsConvertor object */
  if (_array->hasAttribute("CTYPE1"))
      _wcsconv = new WCSConvertor(_array);
  else
      _wcsconv = new WCSConvertor(arrayname);

}

/* Destructor */
ArrayPixelIterator::~ArrayPixelIterator()
{
  PROC;
  
  delete _wcsconv;   _wcsconv=0;
  delete _arraydata; _arraydata=0;
 
  dataSetServer->close(_set); _set = 0;
 
  ENDPROC;
}

/* Iterator methods */
void ArrayPixelIterator::first()
{
  _firstcalled = true;
  _index=0;
  _xindex=1;
  _yindex=1;
}


void ArrayPixelIterator::next ()
{
  assert (!isDone());
  _index++;

  _xindex++;
  if (_xindex > _xbins)
    {
      _xindex=1;
      _yindex++;
    }
}

/* Method to set the index position and start iterating from here */
void ArrayPixelIterator::setIndex (unsigned long index)
{
  _index=index;

  _yindex=int(_index / _xbins) + 1;
  _xindex=_index + 1 - (_yindex - 1) * _xbins;

}

bool ArrayPixelIterator::isDone()
{
  PROC;
  assert(_firstcalled);
  RETURN (_index >= _size);
}


PixelElement ArrayPixelIterator::currentItem()
{
  PROC;

  assert(_firstcalled);

  /* Convert x and y positions to detector coordinates, 
     using WCS attributes */
  /* [HS] now usings new WCSConvertor interface */
  double value = _arrayvalue.asDouble(_index);
  CoordPair physcoord = _wcsconv->physicalCoords(_xindex, _yindex);
  PixelElement result (_xindex,
		       _yindex,
		       physcoord.x(),
		       physcoord.y(),
		       _wcsconv->xWidth(),
		       _wcsconv->yWidth(),
		      value);

  RETURN result;
}

double ArrayPixelIterator::valueAt(unsigned long index) 
{
  PROC;
  RETURN _arrayvalue.asDouble(index);
}

double ArrayPixelIterator::valueAt() 
{
  PROC;
  RETURN _arrayvalue.asDouble(_index);
}

unsigned ArrayPixelIterator::size() const
{
  PROC;
  RETURN _size;
}

unsigned ArrayPixelIterator::xbins() const
{
  PROC;
  RETURN _xbins;
}

unsigned ArrayPixelIterator::ybins() const
{
  PROC;
  RETURN _size / _xbins;
}

/* WCSConvertor object */
const WCSConvertor * ArrayPixelIterator::wcsconvertor() const
{
  PROC;
  RETURN _wcsconv;
}


