// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// This  skeleton file was automatically created by 
//            pkgmaker 0.1
//            Thursday, December 11 1997
// 
// Module: CopyUtils
// Author: Hassan Siddiqui
//
// Description: Namespace for functions to add columns to tables, and tables to datasets.
// 
// Comments: 
//  
// Modifications: See ChangeLog
// 
// $Id: CopyUtils.h,v 1.1 2012/01/25 15:46:31 rsaxton Exp $


#ifndef COPYUTILS_H
#define COPYUTILS_H

// General classes
#include <string> // put this up at the top as order matters
#include <iostream>
#include <cmath>
#include <list>
#include<cassert>

// SAS-specific classes
#include "Dal.h"

// Utility functions
#include "errstr.h"
#include "Tracer.h"

// Macros to help with handling various data types
#include "GenericOperation.h"

#define COPYCOLUMNDATA(TYPE)\
TYPE* fromData = fromColumn ->data() -> TYPE ## Data();\
TYPE* toData   = toColumn   ->data() -> TYPE ## Data();\
for (unsigned long index = 0; index < (toColumn->elements() * toColumn->rows()); ++index)\
{ toData[index] = fromData[index];}

#define COPYARRAYDATA(TYPE)\
TYPE* fromData = fromArray ->data() -> TYPE ## Data();\
TYPE* toData   = toArray   ->data() -> TYPE ## Data();\
for (unsigned long index = 0; index < toArray->elements(); ++index)\
{ toData[index] = fromData[index];}

// -----------------------------------------------------------------------------------

// class CopyUtils

// Name space for copy utility functions

// -----------------------------------------------------------------------------------


class CopyUtils
{

public:

  // Provide convenience functions for copying structures, retaining their original names,
  // and assigning new names, respectively.
  // Overloaded functions are used instead of using the default mechanism
  // as the latter would require the argument to be checked for a unique (the default) value.

  static Column* copyColumnToTable(const Column* fromColumn, 
					 Table* toTable)
    {
      PROC;
      RETURN columnToTable(fromColumn, toTable, fromColumn->name());
    }

  static Column* copyColumnToTable(const Column* fromColumn, 
					 Table* toTable, 
					 const string& newColumnName)
    {
      PROC;
      RETURN columnToTable(fromColumn, toTable, newColumnName);
    }


  static Table* copyTableToDataSet(const Table* fromTable, 
				 DataSet* toSet)
    {
      PROC;
      RETURN tableToDataSet(fromTable, toSet, fromTable->name());
    }

  static Table* copyTableToDataSet(const Table* fromTable, 
				 DataSet* toSet, 
				 const string& newTableName)
    {
      PROC;
      RETURN tableToDataSet(fromTable, toSet, newTableName);
    }



  static Array* copyArrayToDataSet(const Array* fromArray, 
				 DataSet* toSet)
    {
      PROC;
      RETURN arrayToDataSet(fromArray, toSet, fromArray->name());
    }

  static Array* copyArrayToDataSet(const Array* fromArray, 
				 DataSet* toSet, 
				 const string& newArrayName)
    {
      PROC;
      RETURN arrayToDataSet(fromArray, toSet, newArrayName);
    }



private:
  
  static Column* columnToTable(const Column* fromColumn, Table* toTable, const string toColumnName)
    {
      PROC;

      // Check if column of that name already exists in table

      if (toTable->hasColumn(toColumnName))
	{
	  errstr << "Table already has a column named" << toColumnName
		 << Sas::error("duplicateColumnName");

	};

      // Check the number of rows in the column. 
      // Exit on error if it's not compatible with the table.
    
      if (fromColumn->rows() != toTable->rows())
	{
	  errstr << "The number of rows in specified column is " << fromColumn->rows()
		 << " which is incompatible with number in table (" << toTable->rows() 
		 << Sas::error("incompatibleRows");
	}
      

      // Add a new column to the target table. Don't worry about the positioning.

      Column* toColumn = toTable->addColumn(toColumnName, 
					    fromColumn->data()->dataType(),
					    fromColumn->label(),
					    fromColumn->units(),
					    fromColumn->data()->size());


      // Fill the column with data, depending on the data type

      assert(fromColumn->elements() == toColumn->elements());
      assert(fromColumn->rows() == toColumn->rows());

      // See GenericOperation.h 
      // and the preprocessor macro COPYCOLUMNDATA for information on
      // the following expansion.
      GENERIC_OPERATION(COPYCOLUMNDATA, fromColumn->data()->dataType()); 

      // Return a pointer to the new column

      RETURN toColumn;
    }


  static Table* tableToDataSet(const Table* fromTable, DataSet* toSet, const string& toTableName)
    {
      PROC;
      
      // Check if table of that name already exists in dataset
      // Assume that no two blocks of the same name are supported by the DAL.

      if (toSet->hasBlock(toTableName))
	{
	  errstr << "Dataset already has a block named" << toTableName
		 << Sas::error("duplicateBlockName");

	}


      // Add the new table to the target dataset. Don't specify the positioning.

      Table* toTable = toSet->addTable(toTableName, fromTable->rows());


      // Copy attributes over.

      (Attributable&)*toTable = (*fromTable);

      // Copy columns over

      for (unsigned int index = 0; index < fromTable->columns(); ++index)
	{
	  copyColumnToTable(fromTable->column(index /*, AccessMode::Read */), toTable);
	}
      

      // Return new table
      RETURN toTable; 
    }


  static Array* arrayToDataSet(const Array* fromArray, 
				     DataSet* toSet,
				     const string& toArrayName)
    {
      PROC;
      
      // Check if Array of that name already exists in dataset
      // Assume that no two blocks of the same name are supported by the DAL.

      if (toSet->hasBlock(toArrayName))
	{
	  errstr << "Dataset already has a block named" << toArrayName
		 << Sas::error("duplicateBlockName");

	}


      // Add the new Array to the target dataset. Don't specify the positioning.

      Array* toArray = toSet->addArray(toArrayName, 
				       fromArray->data()->dataType(),
				       fromArray->data()->size());


      // Copy attributes over.

      (Attributable&)*toArray = (*fromArray);

      // Copy data over
      
      assert(fromArray->elements() == toArray->elements());

      GENERIC_OPERATION(COPYARRAYDATA, fromArray->data()->dataType());

      RETURN toArray; 
    }

};


#endif
